

#include <cstdlib>
#include <cstdio>

#include <GL/glut.h>
#include "global.h"
#include "ballscuda.h"

const int THREADSPERBLOCK = 64;
const int BLOCKS = N/THREADSPERBLOCK + (N % THREADSPERBLOCK == 0 ? 0 : 1);


//__constant__ float BIGG = 0.002f;
//__constant__ float EPS2 = 0.001;

__constant__ float C = 1.5;
__constant__ float KFG = 200.0;
__constant__ float bkr = 1900;
__constant__ float bkd = 10;

__device__ int selected_ball_gpu_i=-1;

__device__ float4 pos2[N];
__device__ float4 vel2[N];


//__constant__ float3 Gravity = {-5,-5.3,0};

__shared__ float4 shPosition[THREADSPERBLOCK];
__shared__ float4 shVelocity[THREADSPERBLOCK];
__shared__ float shR[THREADSPERBLOCK];
__shared__ float shMass[THREADSPERBLOCK];






__device__ float3 bodybodyinteraction(float4 posi, float4 posj, float4 veli, float4 velj, float Ri, float Rj, float massi, float massj, float3 fi)
{
	float ksi, dksidt, fn;
	float nx,ny,vx,vy;
	float dist;
	float oneoverdist;

	dist = sqrt((posj.x - posi.x)*(posj.x - posi.x) + (posj.y - posi.y)*(posj.y - posi.y));

	if(dist != 0)
	{

		// nbody

		//nx = posj.x - posi.x;
		//ny = posj.y - posi.y;
		//dist2 = sqrt(nx * nx + ny * ny + EPS2);
		//forcidx.x += __fdividef(nx * BIGG , dist2*dist2*dist2);
		//forcidx.y += __fdividef(ny * BIGG , dist2*dist2*dist2);

		ksi =  (Ri + Rj) - dist;

		if ( ksi > 0 )
		{
			oneoverdist = 1.0f/dist;

			//Bell et.al., 'Particle-based simulation of granular materials', equations 1-8, p. 3.

			nx = (posj.x - posi.x) * (oneoverdist);
			ny = (posj.y - posi.y) * (oneoverdist);

			vx = veli.x - velj.x;
			vy = veli.y - velj.y;

			dksidt = vx * nx + vy * ny;

			fn = - bkr * ksi - bkd * dksidt;

			fi.x += fn * nx;
			fi.y += fn * ny;
		}

	}


	return fi;
}


__device__ float3 tile_calculation(float4 pos, float4 vel, float3 forc, float R, float mass)
{
	int i;

	#pragma unroll 64
	for (i = 0; i < blockDim.x; i++)
		forc = bodybodyinteraction(pos, shPosition[i], vel, shVelocity[i], R, shR[i], mass, shMass[i], forc);
		
	return forc;
}



// using shared memory
__global__ void moveparticlesSharedOnDevice(FLOATVEC* posvbo, FLOATVEC *vel, float *R, float *mass, float _DT, float3 _G, int _N)
{
	
	int idx = blockIdx.x*blockDim.x + threadIdx.x;

	FLOATVEC posidx;
	FLOATVEC velidx;
	float Ridx, massidx;
	
	float3 forc;
	float v,dtm;

	int i,tile;

	if (idx<_N) 
	{
		posidx	= posvbo[idx];
		velidx	= vel[idx];
		Ridx	= R[idx];
		massidx	= mass[idx];


		// clear force
		//for(i=0; i < _N; i++)
	
		forc.x = 0;
		forc.y = 0;

		tile = 0;

		for (i = 0; i+threadIdx.x < _N; i += blockDim.x)				// i+threadIdx.x - poprawka do kodu nVidii
		{

			//=========== PARALLEL START HERE ============

			int pair_id = tile * blockDim.x + threadIdx.x;
			
			shPosition[threadIdx.x] = posvbo[pair_id];
			shVelocity[threadIdx.x] =    vel[pair_id];
				   shR[threadIdx.x] =      R[pair_id];
				shMass[threadIdx.x] =   mass[pair_id];

			__syncthreads(); // fill out shPosition & shVelocity

			//=========== PARALLEL END HERE ==============


			//=========== PARALLEL START HERE ============

			forc = tile_calculation(posidx, velidx, forc, Ridx, massidx);
			__syncthreads();	// calculate force acting on 

			//=========== PARALLEL END HERE ==============

			tile++;
		}



		// gravity
		forc.x += _G.x * massidx;
		forc.y += _G.y * massidx;

		// collision walls
		if (posidx.y - Ridx < -0.95)	forc.y = forc.y - KFG * (posidx.y - Ridx - (-0.95));
		if (posidx.x - Ridx < -1.0)		forc.x = forc.x - KFG * (posidx.x - Ridx - (-1.0));
		if (posidx.x + Ridx >  1.0)		forc.x = forc.x - KFG * (posidx.x + Ridx - 1.0);


		if(idx != selected_ball_gpu_i)
		{
			// damp
			//for(int i=0; i<N; i++)
			
			v = sqrt(velidx.x * velidx.x + velidx.y * velidx.y);

			if(v)
			{
				forc.x = forc.x - 2*Ridx * v * velidx.x * (1.0/v) * C;
				forc.y = forc.y - 2*Ridx * v * velidx.y * (1.0/v) * C;
			}

			dtm = _DT / massidx;

			velidx.x = velidx.x + ( dtm ) * forc.x;
			velidx.y = velidx.y + ( dtm ) * forc.y;

			posidx.x = posidx.x + _DT * velidx.x;
			posidx.y = posidx.y + _DT * velidx.y;		
		}

		pos2[idx] = posidx;
		vel2[idx] = velidx;

	}

}







__global__ void swapBuffersOnDevice(FLOATVEC* posvbo, FLOATVEC *vel, int _N)
{
	int idx = blockIdx.x * blockDim.x + threadIdx.x;

	if(idx < _N)
	{
			posvbo[idx] =	pos2[idx];
			vel[idx]	=	vel2[idx];
	}
}

__device__ float3 operator+(const float3 &a, const float3 &b) 
{
  return make_float3(a.x+b.x, a.y+b.y, a.z+b.z);
}

__global__ void moveparticlesOnDevice(FLOATVEC* posvbo, FLOATVEC *vel, float *R, float *mass, float _DT, float3 _G, int _N)
{
	int idx = blockIdx.x * blockDim.x + threadIdx.x;

	float dtm,v;
	float ksi, dksidt, fn;
	float nx,ny,vx,vy;
	float dist,oneoverdist;
	int j;
//	float3 _G = make_float3(0.5,-0.5,0.0);
//	float3 _G=Gravity;


//	float3 g1=make_float3(0,-4,0);
//	float3 g2=make_float3(-4,0,0);
//	float3 _G = g1 + g2;

	float Ridx;
	float massidx;

	float dist2;

	FLOATVEC posidx;
	FLOATVEC velidx;
	FLOATVEC forcidx;

	FLOATVEC posjdx;
	FLOATVEC veljdx;

	if(idx < _N)
	{

		posidx = posvbo[idx];
		velidx = vel[idx];
		Ridx = R[idx];
		massidx = mass[idx];


		// gravity
		forcidx.x = _G.x * massidx;
		forcidx.y = _G.y * massidx;

		// collision walls
		if (posidx.y - Ridx < -0.95)	forcidx.y = forcidx.y - KFG * (posidx.y - Ridx - (-0.95));
		if (posidx.x - Ridx < -1.0)		forcidx.x = forcidx.x - KFG * (posidx.x - Ridx - (-1.0));
		if (posidx.x + Ridx >  1.0)		forcidx.x = forcidx.x - KFG * (posidx.x + Ridx - 1.0);



		for(j=0; j<_N; j++)
		{
			posjdx = posvbo[j];

			dist = sqrt((posjdx.x - posidx.x)*(posjdx.x - posidx.x) + (posjdx.y - posidx.y)*(posjdx.y - posidx.y));

			if(dist != 0)
			{
	
				// nbody

				//nx = posjdx.x - posidx.x;
				//ny = posjdx.y - posidx.y;
				//dist2 = sqrt(nx * nx + ny * ny + EPS2);
				//forcidx.x += __fdividef(nx * BIGG , dist2*dist2*dist2);
				//forcidx.y += __fdividef(ny * BIGG , dist2*dist2*dist2);

				ksi =  (Ridx + R[j]) - dist;

				if ( ksi > 0 )
				{
					oneoverdist = 1.0f/dist;

					//Bell et.al., 'Particle-based simulation of granular materials', equations 1-8, p. 3.

					nx = (posjdx.x - posidx.x) * (oneoverdist);
					ny = (posjdx.y - posidx.y) * (oneoverdist);

					veljdx = vel[j];
			
					vx = velidx.x - veljdx.x;
					vy = velidx.y - veljdx.y;

					dksidt = vx * nx + vy * ny;

					fn = - bkr * ksi - bkd * dksidt;

					forcidx.x = forcidx.x + fn * nx;
					forcidx.y = forcidx.y + fn * ny;
				}

			}
		}


		if(idx != selected_ball_gpu_i)
		{
			// damp
			//for(int i=0; i<N; i++)
			
			v = sqrt(velidx.x * velidx.x + velidx.y * velidx.y);

			if(v)
			{
				forcidx.x = forcidx.x - 2*Ridx * v * velidx.x * (1.0/v) * C;
				forcidx.y = forcidx.y - 2*Ridx * v * velidx.y * (1.0/v) * C;
			}

			dtm = _DT / massidx;

			velidx.x = velidx.x + ( dtm ) * forcidx.x;
			velidx.y = velidx.y + ( dtm ) * forcidx.y;

			posidx.x = posidx.x + _DT * velidx.x;
			posidx.y = posidx.y + _DT * velidx.y;		
		}

		pos2[idx] = posidx;
		vel2[idx] = velidx;
	}
}


FLOATVEC *vel_device;
float *mass_device;
float *R_device;
//FLOATVEC *selected_ball_r_device;


FLOATVEC *pos_host;
FLOATVEC *vel_host;
FLOATVEC *forc_host;
float *mas_host;
float *R_host;

void allocate_memory_cuda(int _N)
{
	printf("alokacja... \n");

	size_t size = _N*sizeof(FLOATVEC);
	size_t sizefloat = _N*sizeof(float);

	pos_host = (FLOATVEC *) malloc(size);
	vel_host = (FLOATVEC *) malloc(size);
	forc_host = (FLOATVEC *) malloc(size);
	mas_host = (float *) malloc(sizefloat);
	R_host = (float *) malloc(sizefloat);
	

	cudaMalloc((void **) &vel_device, size);
	cudaMalloc((void **) &mass_device, sizefloat);
	cudaMalloc((void **) &R_device, sizefloat);
}

void deallocate_memory_cuda(void)
{
	printf("dealokacja... \n");

	free(pos_host);
	free(vel_host);
	free(forc_host);
	free(mas_host);
	free(R_host);
	
	cudaFree(vel_device); 
	cudaFree(mass_device); 
	cudaFree(R_device); 
}

void copy_memory_cuda(int _N)
{
	cudaMemcpy(vel_device, vel_host, sizeof(float4)*_N, cudaMemcpyHostToDevice);
	cudaMemcpy(mass_device, mas_host, sizeof(float)*_N, cudaMemcpyHostToDevice);
	cudaMemcpy(R_device, R_host, sizeof(float)*_N, cudaMemcpyHostToDevice);
}


__global__ void findselectOnDevice(FLOATVEC* posvbo, float *R, FLOATVEC pos, int _N)
{
	int idx = blockIdx.x * blockDim.x + threadIdx.x;
	float dist;
	FLOATVEC posidx;

	if(idx < _N)
	{
		posidx = posvbo[idx];

		dist = sqrt (	(pos.x - posidx.x) * (pos.x - posidx.x) +
						(pos.y - posidx.y) * (pos.y - posidx.y));
		
		if(dist - R[idx] < 0)
		{	
			selected_ball_gpu_i = idx;

			posvbo[selected_ball_gpu_i].x = pos.x;
			posvbo[selected_ball_gpu_i].y = pos.y;
		}
	}
}

__global__ void changeposOnDevice(FLOATVEC* posvbo, FLOATVEC pos)
{
	posvbo[selected_ball_gpu_i].x = pos.x;
	posvbo[selected_ball_gpu_i].y = pos.y;
}




__device__ float4 poz_sel;
__device__ float4 poz_0;
__global__ void read_balls_positions_OnDevice(FLOATVEC* posvbo)
{
	poz_0 = posvbo[0];
	if( selected_ball_gpu_i >= 0)
		poz_sel = posvbo[ selected_ball_gpu_i ];
}


// reads position of selected ball
void readposCUDA(FLOATVEC* posvbo, int _N, float4 *pozsel, float4 *poz0)
{

	read_balls_positions_OnDevice <<< 1, 1 >>> (posvbo);
	cudaThreadSynchronize();

	const char *symbol="poz_sel";
	cudaMemcpyFromSymbol (pozsel, symbol, sizeof(FLOATVEC), 0, cudaMemcpyDeviceToHost);
	const char *symbol0="poz_0";
	cudaMemcpyFromSymbol (poz0, symbol0, sizeof(FLOATVEC), 0, cudaMemcpyDeviceToHost);

}

void findselectCUDA(FLOATVEC* posvbo, FLOATVEC pos, int _N)
{
	findselectOnDevice <<< BLOCKS, THREADSPERBLOCK >>> (posvbo, R_device, pos, _N);
}


void changePosCUDA(FLOATVEC* posvbo, FLOATVEC pos)
{
	changeposOnDevice <<< 1, 1 >>> (posvbo, pos);
}

void moveParticlesCUDA(FLOATVEC* posvbo, float _DT, float3 _G, int _N)
{
	//moveparticlesOnDevice <<< BLOCKS, THREADSPERBLOCK >>> (posvbo, vel_device, R_device, mass_device, _DT, _G, _N);
	moveparticlesSharedOnDevice <<< BLOCKS, THREADSPERBLOCK >>> (posvbo, vel_device, R_device, mass_device, _DT, _G, _N);
	//cudaThreadSynchronize();
	swapBuffersOnDevice <<< BLOCKS, THREADSPERBLOCK >>> (posvbo, vel_device, _N);
	//cudaThreadSynchronize();
}
